/*
Copyright (C) 2007-2018 Victor Matei Petrescu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "grconfig.h"
#include "graphics.h"

#if SDLVER==0
  #undef SOUND
  #define SOUND 0
#endif

#if SOUND == 1

extern SDL_AudioSpec *RGLOB_desired, *RGLOB_obtained;
extern float RGLOB_userdata[20];
float RGLOB_x_sin[4]={0.0f,0.0f,0.0f,0.0f};

float sinrd(){
  unsigned int r;
  float x;
r=rand(); r&=255;
x=0.00784f*r-1.0f;
return x;
}

/*audio callback function for SDL*/
void my_audio_callback(void *userdata, Uint8 *stream, int len){
  int i,j,freq;
  float dx[4],pi2,pi2f,a,*usrd;
  Uint16 u16,format;
  Uint8 u8,u81,channels;
  Sint16 s16;
  Sint8 s8;

format=(Uint16)RGLOB_obtained->format;
channels=(Uint8)RGLOB_obtained->channels;
if((channels!=1)&&(channels!=2)){
  printf("my_audio_callback(): Unsupported number of channels\r\n"); exit(1);
}
freq=(int)RGLOB_obtained->freq;
usrd=(float *)userdata;
pi2=6.283185f;
pi2f=pi2/freq;

for(i=0;i<4;i++){dx[i]=pi2f*usrd[i<<1];}

i=0;
while(i<len){
  a=0.0f;
  for(j=0;j<3;j++){
    RGLOB_x_sin[j]+=dx[j];
    if(RGLOB_x_sin[j]>pi2){RGLOB_x_sin[j]-=pi2;}
    a+=0.25f*usrd[(j<<1)+1]*sin(RGLOB_x_sin[j]);
  }
  a+=0.25f*usrd[7]*sinrd();

  switch(format){
    case AUDIO_U8:
      u8=(Uint8)(127+a*127); stream[i]=u8; i++;
      if(channels==2){stream[i]=u8; i++;}
      break;
    case AUDIO_S8:
      s8=(Sint8)(a*127);
      if(s8<0){u8=~(-s8);}else{u8=s8;}
      stream[i]=u8; i++;
      if(channels==2){stream[i]=u8; i++;}
      break;
    case AUDIO_U16:
      u16=(Uint16)(32767+a*32767);
      u8=(u16>>8) & 255; u81=u16 & 255;
      stream[i]=u81; i++; stream[i]=u8; i++;
      if(channels==2){stream[i]=u81; i++; stream[i]=u8; i++;}
      break;
    case AUDIO_S16:
      s16=(Sint16)(a*32767);
      if(s16<0){
        s16=-s16;
        u8=(s16>>8) & 255; u8=~u8;
        u81=s16 & 255; u81=~u81;
      }else{
        u8=(s16>>8) & 255;
        u81=s16 & 255;
      }
      stream[i]=u81; i++; stream[i]=u8; i++;
      if(channels==2){stream[i]=u81; i++; stream[i]=u8; i++;}
      break;
    default: i+=256; break;
  }
}

}

void setvoice(unsigned int v,float freq,float vol){
  if(v<4){
    if(freq<10.0f){freq=10.0f;} /*min.10Hz*/
    if(freq>5000.0f){freq=5000.0f;} /*max.5000Hz*/
    if(vol<0.0f){vol=0.0f;}
    if(vol>1.0f){vol=1.0f;}
    RGLOB_userdata[v<<1]=freq;
    RGLOB_userdata[(v<<1)+1]=vol;
  }else{printf("setvoice(): Voice %d not exist\r\n",v);}
}

#else

void setvoice(unsigned int v,float freq,float vol){printf("setvoice() requires '#define SOUND 1'\r\n");}

#endif
